Require Import ssreflect ssrnat seq eqtype.
Require Import mtac hash.


Module WithList.

  Definition dyn := { x : Prop | x}.
  Definition Dyn := @exist Prop (fun p=>p).

  Definition ProofNotFound : Exception.
    exact exception.
  Qed.

  Program
  Definition lookup (p : Prop)  := 
    mfix f [s : list dyn] : T p :=
      mmatch s return T p with
      | [x s'] (Dyn p x) :: s' => ret x
      | [d s'] d :: s' => f s'
      | _ => raise ProofNotFound
      end.
  
  Program
  Definition tauto' :=
    mfix f [c : list dyn; p : Prop] : T p :=
      mmatch p as p' return T p' with
      | True => ret I 
      | [p1 p2] p1 /\ p2 =>
           r1 <- f c p1 ;
           r2 <- f c p2 ;
           ret (conj r1 r2)
      | [p1 p2]  p1 \/ p2 =>
           mtry 
             r1 <- f c p1 ;
             ret (or_introl r1)
           with _ =>
             r2 <- f c p2 ;
             ret (or_intror r2)
           end
      | [p1 p2 : Prop] p1 -> p2 =>
           nu (x:p1),
             r <- f (Dyn p1 x :: c) p2;
             abs x r
      | [A (q:A -> Prop)] (forall x:A, q x) =>
           nu (x:A),
             r <- f c (q x);
             abs x r
      | [A (q:A -> Prop)] (exists x:A, q x) =>
           X <- evar A;
           r <- f c (q X);
           b <- is_evar X;
           if b then 
             raise ProofNotFound
           else
             ret (ex_intro q X r)
      | [p':Prop] p' => lookup p' c
      end.
  
  Definition tauto P := 
    tauto' nil P.

End WithList.

Module WithHash.

  Definition ProofNotFound : Exception.
    exact exception.
  Qed.

  Definition ctx := HashTbl.t Type (fun x=>x).

  Definition empty_ctx : T ctx :=
    HashTbl.create Type (fun x=>x).

  Program
  Definition tauto' (c : ctx) :=
    mfix f [p : Prop] : T p :=
    mmatch p as p' return T p' with
    | True => ret I 
    | [p1 p2] p1 /\ p2 =>
         r1 <- f p1 ;
         r2 <- f p2 ;
         ret (conj r1 r2)
    | [p1 p2]  p1 \/ p2 =>
         mtry 
           r1 <- f p1 ;
           ret (or_introl r1)
         with _ =>
           r2 <- f p2 ;
           ret (or_intror r2)
         end
    | [A (q:A -> Prop)] (forall x:A, q x) =>
          nu (x:A),
          HashTbl.add c A x;;
          r <- f (q x);
          abs x r
    | [x] x => 
      mtry HashTbl.find c x
      with _ => raise ProofNotFound
      end
    end.

  Definition tauto P := 
    c <- empty_ctx;
    tauto' c P.

End WithHash.

Example ex_first_order_0 : 
  forall x (p q : nat -> Prop), exists y, p x -> q x -> p y /\ q y.
Proof. 
  apply: (run (WithList.tauto _)).
Qed.

Example ex_first_order_1 : 
  forall x (p q : nat -> Prop), p x -> q x -> p x /\ q x.
Proof. 
  apply: (run (WithList.tauto _)).
Qed.

Definition test_ref (z : nat) :=
  r <- ref 0;
  (nu x : nat, r ::= z);;
  !r.

Check (fun z => run test_ref z).


Definition test_hash (z : nat) :=
  hz1 <- hash z 10;
  nu x:nat,
    hz2 <- hash z 10;
    hx1 <- hash x 10;
    (nu y:nat,
      hz3 <- hash z 10;
      hx2 <- hash x 10;
      hy1 <- hash y 10;
      (nu w:nat,
        hz4 <- hash z 10;
        hx3 <- hash x 10;
        hy2 <- hash y 10;
        ret ((hz1 :: hz2 :: hz3 :: hz4 :: nil), (hx1 :: hx2 :: hx3 :: nil), (hy1 :: hy2 :: nil)))).
      
Check (fun w=>run test_hash w).


Example ex_first_order_2 (p q r : Prop) : 
  q -> p -> q.
Proof. 
  apply: (run (WithHash.tauto _)).
Qed.

Example ex_first_order_2'  : 
  forall (p q : nat -> Prop) x , p x -> q x -> p x /\ q x.
Proof. 
  apply: (run (WithHash.tauto _)).
Qed.


Example ex_stress_1 
  (p00 p01 p02 p03 p04 p05 p06 p07 p08 p09 : Prop) :
  p00 -> p01 -> p02 -> p03 -> p04 -> p05 -> p06 -> p07 -> p08 -> p09 ->
  p00 /\ p01 /\ p02 /\ p03 /\ p04 /\ p05 /\ p06 /\ p07 /\ p08 /\ p09.
Proof. 
  Time apply: (run (WithHash.tauto _)).
Qed.

Example ex_stress_2
  (p00 p01 p02 p03 p04 p05 p06 p07 p08 p09 : Prop)
  (p10 p11 p12 p13 p14 p15 p16 p17 p18 p19 : Prop) :

  p00 -> p01 -> p02 -> p03 -> p04 -> p05 -> p06 -> p07 -> p08 -> p09 ->
  p10 -> p11 -> p12 -> p13 -> p14 -> p15 -> p16 -> p17 -> p18 -> p19 ->

  p00 /\ p01 /\ p02 /\ p03 /\ p04 /\ p05 /\ p06 /\ p07 /\ p08 /\ p09 /\
  p10 /\ p11 /\ p12 /\ p13 /\ p14 /\ p15 /\ p16 /\ p17 /\ p18 /\ p19.
Proof. 
  Time apply: (run (WithHash.tauto _)).
Qed.

Example ex_stress_3
  (p00 p01 p02 p03 p04 p05 p06 p07 p08 p09 : Prop)
  (p10 p11 p12 p13 p14 p15 p16 p17 p18 p19 : Prop)
  (p20 p21 p22 p23 p24 p25 p26 p27 p28 p29 : Prop) :

  p00 -> p01 -> p02 -> p03 -> p04 -> p05 -> p06 -> p07 -> p08 -> p09 ->
  p10 -> p11 -> p12 -> p13 -> p14 -> p15 -> p16 -> p17 -> p18 -> p19 ->
  p20 -> p21 -> p22 -> p23 -> p24 -> p25 -> p26 -> p27 -> p28 -> p29 ->

  p00 /\ p01 /\ p02 /\ p03 /\ p04 /\ p05 /\ p06 /\ p07 /\ p08 /\ p09 /\
  p10 /\ p11 /\ p12 /\ p13 /\ p14 /\ p15 /\ p16 /\ p17 /\ p18 /\ p19 /\
  p20 /\ p21 /\ p22 /\ p23 /\ p24 /\ p25 /\ p26 /\ p27 /\ p28 /\ p29.
Proof. 
  Time apply: (run (WithHash.tauto _)).
Qed.

Example ex_stress_4
  (p00 p01 p02 p03 p04 p05 p06 p07 p08 p09 : Prop)
  (p10 p11 p12 p13 p14 p15 p16 p17 p18 p19 : Prop)
  (p20 p21 p22 p23 p24 p25 p26 p27 p28 p29 : Prop)
  (p30 p31 p32 p33 p34 p35 p36 p37 p38 p39 : Prop) :

  p00 -> p01 -> p02 -> p03 -> p04 -> p05 -> p06 -> p07 -> p08 -> p09 ->
  p10 -> p11 -> p12 -> p13 -> p14 -> p15 -> p16 -> p17 -> p18 -> p19 ->
  p20 -> p21 -> p22 -> p23 -> p24 -> p25 -> p26 -> p27 -> p28 -> p29 ->
  p30 -> p31 -> p32 -> p33 -> p34 -> p35 -> p36 -> p37 -> p38 -> p39 ->

  p00 /\ p01 /\ p02 /\ p03 /\ p04 /\ p05 /\ p06 /\ p07 /\ p08 /\ p09 /\
  p10 /\ p11 /\ p12 /\ p13 /\ p14 /\ p15 /\ p16 /\ p17 /\ p18 /\ p19 /\
  p20 /\ p21 /\ p22 /\ p23 /\ p24 /\ p25 /\ p26 /\ p27 /\ p28 /\ p29 /\
  p30 /\ p31 /\ p32 /\ p33 /\ p34 /\ p35 /\ p36 /\ p37 /\ p38 /\ p39.
Proof. 
  Time apply: (run (WithHash.tauto _)).
Qed.




Example ex_stress_l1 
  (p00 p01 p02 p03 p04 p05 p06 p07 p08 p09 : Prop) :
  p00 -> p01 -> p02 -> p03 -> p04 -> p05 -> p06 -> p07 -> p08 -> p09 ->
  p00 /\ p01 /\ p02 /\ p03 /\ p04 /\ p05 /\ p06 /\ p07 /\ p08 /\ p09.
Proof. 
  Time apply: (run (WithList.tauto _)).
Qed.

Example ex_stress_l2
  (p00 p01 p02 p03 p04 p05 p06 p07 p08 p09 : Prop)
  (p10 p11 p12 p13 p14 p15 p16 p17 p18 p19 : Prop) :

  p00 -> p01 -> p02 -> p03 -> p04 -> p05 -> p06 -> p07 -> p08 -> p09 ->
  p10 -> p11 -> p12 -> p13 -> p14 -> p15 -> p16 -> p17 -> p18 -> p19 ->

  p00 /\ p01 /\ p02 /\ p03 /\ p04 /\ p05 /\ p06 /\ p07 /\ p08 /\ p09 /\
  p10 /\ p11 /\ p12 /\ p13 /\ p14 /\ p15 /\ p16 /\ p17 /\ p18 /\ p19.
Proof. 
  Time apply: (run (WithList.tauto _)).
Qed.

Example ex_stress_l3
  (p00 p01 p02 p03 p04 p05 p06 p07 p08 p09 : Prop)
  (p10 p11 p12 p13 p14 p15 p16 p17 p18 p19 : Prop)
  (p20 p21 p22 p23 p24 p25 p26 p27 p28 p29 : Prop) :

  p00 -> p01 -> p02 -> p03 -> p04 -> p05 -> p06 -> p07 -> p08 -> p09 ->
  p10 -> p11 -> p12 -> p13 -> p14 -> p15 -> p16 -> p17 -> p18 -> p19 ->
  p20 -> p21 -> p22 -> p23 -> p24 -> p25 -> p26 -> p27 -> p28 -> p29 ->

  p00 /\ p01 /\ p02 /\ p03 /\ p04 /\ p05 /\ p06 /\ p07 /\ p08 /\ p09 /\
  p10 /\ p11 /\ p12 /\ p13 /\ p14 /\ p15 /\ p16 /\ p17 /\ p18 /\ p19 /\
  p20 /\ p21 /\ p22 /\ p23 /\ p24 /\ p25 /\ p26 /\ p27 /\ p28 /\ p29.
Proof. 
  Time apply: (run (WithList.tauto _)).
Qed.

Example ex_stress_l4
  (p00 p01 p02 p03 p04 p05 p06 p07 p08 p09 : Prop)
  (p10 p11 p12 p13 p14 p15 p16 p17 p18 p19 : Prop)
  (p20 p21 p22 p23 p24 p25 p26 p27 p28 p29 : Prop)
  (p30 p31 p32 p33 p34 p35 p36 p37 p38 p39 : Prop) :

  p00 -> p01 -> p02 -> p03 -> p04 -> p05 -> p06 -> p07 -> p08 -> p09 ->
  p10 -> p11 -> p12 -> p13 -> p14 -> p15 -> p16 -> p17 -> p18 -> p19 ->
  p20 -> p21 -> p22 -> p23 -> p24 -> p25 -> p26 -> p27 -> p28 -> p29 ->
  p30 -> p31 -> p32 -> p33 -> p34 -> p35 -> p36 -> p37 -> p38 -> p39 ->

  p00 /\ p01 /\ p02 /\ p03 /\ p04 /\ p05 /\ p06 /\ p07 /\ p08 /\ p09 /\
  p10 /\ p11 /\ p12 /\ p13 /\ p14 /\ p15 /\ p16 /\ p17 /\ p18 /\ p19 /\
  p20 /\ p21 /\ p22 /\ p23 /\ p24 /\ p25 /\ p26 /\ p27 /\ p28 /\ p29 /\
  p30 /\ p31 /\ p32 /\ p33 /\ p34 /\ p35 /\ p36 /\ p37 /\ p38 /\ p39.
Proof. 
  Time apply: (run (WithList.tauto _)).
Qed.



Example ex_stress_5
  (p00 p01 p02 p03 p04 p05 p06 p07 p08 p09 : Prop)
  (p10 p11 p12 p13 p14 p15 p16 p17 p18 p19 : Prop)
  (p20 p21 p22 p23 p24 p25 p26 p27 p28 p29 : Prop)
  (p30 p31 p32 p33 p34 p35 p36 p37 p38 p39 : Prop) 
  (p40 p41 p42 p43 p44 p45 p46 p47 p48 p49 : Prop) :

  p00 -> p01 -> p02 -> p03 -> p04 -> p05 -> p06 -> p07 -> p08 -> p09 ->
  p10 -> p11 -> p12 -> p13 -> p14 -> p15 -> p16 -> p17 -> p18 -> p19 ->
  p20 -> p21 -> p22 -> p23 -> p24 -> p25 -> p26 -> p27 -> p28 -> p29 ->
  p30 -> p31 -> p32 -> p33 -> p34 -> p35 -> p36 -> p37 -> p38 -> p39 ->
  p40 -> p41 -> p42 -> p43 -> p44 -> p45 -> p46 -> p47 -> p48 -> p49 ->

  p00 /\ p01 /\ p02 /\ p03 /\ p04 /\ p05 /\ p06 /\ p07 /\ p08 /\ p09 /\
  p10 /\ p11 /\ p12 /\ p13 /\ p14 /\ p15 /\ p16 /\ p17 /\ p18 /\ p19 /\
  p20 /\ p21 /\ p22 /\ p23 /\ p24 /\ p25 /\ p26 /\ p27 /\ p28 /\ p29 /\
  p30 /\ p31 /\ p32 /\ p33 /\ p34 /\ p35 /\ p36 /\ p37 /\ p38 /\ p39 /\
  p40 /\ p41 /\ p42 /\ p43 /\ p44 /\ p45 /\ p46 /\ p47 /\ p48 /\ p49.
Proof. 
  Time apply: (run (WithHash.tauto _)).
Qed.



Example ex_stress_l5
  (p00 p01 p02 p03 p04 p05 p06 p07 p08 p09 : Prop)
  (p10 p11 p12 p13 p14 p15 p16 p17 p18 p19 : Prop)
  (p20 p21 p22 p23 p24 p25 p26 p27 p28 p29 : Prop)
  (p30 p31 p32 p33 p34 p35 p36 p37 p38 p39 : Prop) 
  (p40 p41 p42 p43 p44 p45 p46 p47 p48 p49 : Prop) :

  p00 -> p01 -> p02 -> p03 -> p04 -> p05 -> p06 -> p07 -> p08 -> p09 ->
  p10 -> p11 -> p12 -> p13 -> p14 -> p15 -> p16 -> p17 -> p18 -> p19 ->
  p20 -> p21 -> p22 -> p23 -> p24 -> p25 -> p26 -> p27 -> p28 -> p29 ->
  p30 -> p31 -> p32 -> p33 -> p34 -> p35 -> p36 -> p37 -> p38 -> p39 ->
  p40 -> p41 -> p42 -> p43 -> p44 -> p45 -> p46 -> p47 -> p48 -> p49 ->

  p00 /\ p01 /\ p02 /\ p03 /\ p04 /\ p05 /\ p06 /\ p07 /\ p08 /\ p09 /\
  p10 /\ p11 /\ p12 /\ p13 /\ p14 /\ p15 /\ p16 /\ p17 /\ p18 /\ p19 /\
  p20 /\ p21 /\ p22 /\ p23 /\ p24 /\ p25 /\ p26 /\ p27 /\ p28 /\ p29 /\
  p30 /\ p31 /\ p32 /\ p33 /\ p34 /\ p35 /\ p36 /\ p37 /\ p38 /\ p39 /\
  p40 /\ p41 /\ p42 /\ p43 /\ p44 /\ p45 /\ p46 /\ p47 /\ p48 /\ p49.
Proof. 
  Time apply: (run (WithList.tauto _)).
Qed.
